/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * rules_int.c                                               *
 * part of checkers project, contains functions describing   *
 * rules of international checkers                           *
 * Miroslav Tynovsky 02/2004                                 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

//TODO: dopsat ke vsem read-only parametrum const!
#include <stdlib.h>
#include <assert.h>
#include "rules_int.h"

#define MIN_OCCURED 20 /* min. allocated memory for occured positions */

#define is_white_piece(x) ((x)>0)
#define is_black_piece(x) ((x)<0)
#define is_out_of_board(x, y) ((x)>=BOARDSIZE || (x)<0 || (y)>=BOARDSIZE || (y)<0)
#define is_promoting(y, color) ( (((y)==0)&&((color)==BLACK))\
                              || (((y)==BOARDSIZE-1)&&((color)==WHITE)) )
#define correct_coords(x, y) ((!is_out_of_board((x), (y))) && (is_black_square((x), (y))))

/* LOCAL FUNCION PROTOTYPES: */
Game_state manage_occured(Position *position, 
                          Occured_position *occured_positions,
                          Move *last_move);
void make_blockmove(Move *block);
void add_moves(Moves *moves, Moves moves1);
Moves extend_jump(Move first_part, Moves extensions);
Moves king_jump(Direction direction, Position *position, int start_x, int start_y);
Moves king_step(Direction direction, Position *position, int start_x, int start_y);
Moves pawn_jump(Direction direction, Position *position, int start_x, int start_y);
Moves pawn_step(Direction direction, Position *position, int start_x, int start_y);
Moves pawn_steps(Position *position, int x, int y);
Moves pawn_jumps(Position *position, int x, int y);
Moves king_jumps(Position *position, int x, int y);
Moves king_steps(Position *position, int x, int y);
Moves get_pawn_moves(Position *position, int x, int y);
Moves get_king_moves(Position *position, int x, int y);
Moves get_piece_moves(Position *position, int x, int y);
int color(int x);

/* GLOBAL FUNCTIONS */

void initialize_board(Position *position)
{
	int i, j;

	for(i=0; i<BOARDSIZE; i++)
		for(j=0; j<BOARDSIZE; j++)
			position->placement[i][j]= NONE;

	/* first 4 rows fill with white pawns */
	for(i=0; i<BOARDSIZE; i++) 
		for(j=0; j<4; j++)
			if(is_black_square(i,j)) 
				position->placement[i][j]= WHITE_PAWN;

	/* last 4 rows fill with black pawns */
	for(i=0; i<BOARDSIZE; i++) 
		for(j=BOARDSIZE-4; j<BOARDSIZE; j++)
			if(is_black_square(i,j)) 
				position->placement[i][j]= BLACK_PAWN;

	position->onmove= WHITE;
}

/* get_game_state() returns wheather black/white wins or 
 * it is draw or the game is still proceeding.
 */
Game_state get_game_state(const Position *position, 
                          Occured_position *occured_positions, 
                          Move *last_move)
{
	Moves moves;
	int moves_count;

	moves= get_all_moves(position);
	moves_count= moves.count;
	if(moves.allocated > 0) free(moves.list);

	if(moves_count==0){
		switch(position->onmove){
			case WHITE: return BLACK_WINS;
			case BLACK: return WHITE_WINS;
		}
	}
	return PROCEEDING;//manage_occured(position, occured_positions, last_move);
}

/* get_all_moves() returns array of all possible moves in given position
 */
Moves get_all_moves(const Position *position)
{
	int x, y;
	Moves moves;
	moves.list= NULL;
	moves.allocated= 0;
	moves.count= 0;

	for(y=0; y<BOARDSIZE; y++)  //goes by rows :)
		for(x=0; x<BOARDSIZE; x++)
			if(is_black_square(x,y))
				add_moves(&moves, get_piece_moves(position, x, y));
	return moves;
}

/* do_move() changes position according to execution of given move 
 */
void do_move(Position *position, Move move)
{
	int i;

	assert(correct_coords(move.destination.x, move.destination.y));
	assert(correct_coords(move.start.x, move.start.y));

	switch(move.move_type){
	case PAWN_STEP:
	case KING_STEP:
		position->placement[move.destination.x][move.destination.y]
			= position->placement[move.start.x][move.start.y];
		position->placement[move.start.x][move.start.y]= NONE;
		break;

	case PAWN_JUMP:
	case KING_JUMP:
		position->placement[move.destination.x][move.destination.y]
			= position->placement[move.start.x][move.start.y];
		if((move.start.x != move.destination.x) || (move.start.y != move.destination.y))
		position->placement[move.start.x][move.start.y]= NONE;
		/* clean captured pieces: */
		for(i=0; i<move.jump_length; i++)
			if(    (move.captured_coords[i].x != move.destination.x) 
			    || (move.captured_coords[i].y != move.destination.y)  )
				position->placement[move.captured_coords[i].x][move.captured_coords[i].y]= NONE;
		break;

	case PROMOTION:
		position->placement[move.destination.x][move.destination.y]
			= promote(position->placement[move.start.x][move.start.y]);
		position->placement[move.start.x][move.start.y]= NONE;
		break;

	case JUMP_PROMOTION:
		position->placement[move.destination.x][move.destination.y]
			= promote(position->placement[move.start.x][move.start.y]);
		position->placement[move.start.x][move.start.y]= NONE;
		/* clean captured pieces: */
		for(i=0; i<move.jump_length; i++)
			position->placement[move.captured_coords[i].x][move.captured_coords[i].y]= NONE;
		break;
	}
	position->onmove= -position->onmove;
}

/* undo_move() changes position according to taking back of given move 
 */
void undo_move(Position *position, Move move)
{
	int i;

	assert(correct_coords(move.destination.x, move.destination.y));
	assert(correct_coords(move.start.x, move.start.y));

	switch(move.move_type){
	case PAWN_STEP:
	case KING_STEP:
		position->placement[move.start.x][move.start.y]
			= position->placement[move.destination.x][move.destination.y];
		position->placement[move.destination.x][move.destination.y]= NONE;
		break;

	case PAWN_JUMP:
	case KING_JUMP:
		position->placement[move.start.x][move.start.y]
			= position->placement[move.destination.x][move.destination.y];
		if((move.start.x != move.destination.x) || (move.start.y != move.destination.y))
			position->placement[move.destination.x][move.destination.y]= NONE;
		/* give back captured pieces: */
		for(i=0; i<move.jump_length; i++)
			position->placement[move.captured_coords[i].x][move.captured_coords[i].y]=
				move.captured_pieces[i];
		break;

	case PROMOTION:
		position->placement[move.start.x][move.start.y]
			= degrade(position->placement[move.destination.x][move.destination.y]);
		position->placement[move.destination.x][move.destination.y]= NONE;
		break;

	case JUMP_PROMOTION:
		position->placement[move.start.x][move.start.y]
			= degrade(position->placement[move.destination.x][move.destination.y]);
		position->placement[move.destination.x][move.destination.y]= NONE;
		/* give back captured pieces: */
		for(i=0; i<move.jump_length; i++)
			position->placement[move.captured_coords[i].x][move.captured_coords[i].y]= 
				move.captured_pieces[i];
		break;
	} /* end of switch */
	position->onmove= -position->onmove;
} /* end of undo_move() */


/* returns moves from `moves' that start at sq_start. frees list of `moves'.
 */
Moves select_moves_by_start(Sq_coords sq_start, Moves moves)
{
	Moves selected;
	Moves one_move;
	int i;
/* tady to pada */
	assert(correct_coords(sq_start.x, sq_start.y));
	
	selected.list= (Move *)malloc(4*sizeof(Move));
	selected.count= 0;
	selected.allocated= 4;

	for(i=0; i<moves.count; i++){
		if ( (moves.list[i].start.x == sq_start.x)
		     && (moves.list[i].start.y == sq_start.y) ){
			one_move.list= (Move *)malloc(sizeof(Move));
			one_move.count= 1;
			one_move.allocated= 1;
			one_move.list[0]= moves.list[i];
			add_moves(&selected, one_move);
		}
	}
	free(moves.list);
	return selected;
}

/* returns 1 if there is a move in `moves' starting with sq_start -> sq_dest */

int is_member(Sq_coords sq_start, Sq_coords sq_dest, Moves moves)
{
	int i, jump_length;

	for(i=0; i<moves.count; i++){
		jump_length= moves.list[i].jump_length;
		if(jump_length>1){
			if( (moves.list[i].start.x == sq_start.x)
			   && (moves.list[i].start.y == sq_start.y)
			   && (moves.list[i].middle[jump_length-2].x == sq_dest.x)
			   && (moves.list[i].middle[jump_length-2].y == sq_dest.y) ){
				return i;
			}
		}else{
			if( (moves.list[i].start.x == sq_start.x)
			   && (moves.list[i].start.y == sq_start.y)
			   && (moves.list[i].destination.x == sq_dest.x)
			   && (moves.list[i].destination.y == sq_dest.y) ){
				return i;
			} 
		}
	}
	return -1;
}

/* returns a move from the `moves' which is equal to given `move' in start, 
 * middle and destination
 */

Move match_move(Move move, const Moves *moves)
{
	int i, j;

	for(i=0; i<moves->count; i++){
		if(  (moves->list[i].start.x==move.start.x)
		  && (moves->list[i].start.y==move.start.y)
		  && (moves->list[i].destination.x==move.destination.x)
		  && (moves->list[i].destination.y==move.destination.y) ){
			for(j=0; j < moves->list[i].jump_length; j++)
				if( (move.middle[j].x!=moves->list[i].middle[j].x)
				  ||(move.middle[j].y!=moves->list[i].middle[j].y) )
					break;
			return moves->list[i];
		}
	}
	move.start.x= -1;
	return move;
}


/* eq_positions() returns 1 if given positions are equal, 0 if they are not. 
 */
int eq_positions(Position position1, Position position2)
{
	if(position1.onmove != position2.onmove){
		return 0;
	}else{
		int i,j;
		for(i=0; i<BOARDSIZE; i++)
			for(j=0; j<BOARDSIZE; j++)
				if(position1.placement[i][j] != position2.placement[i][j])
					return 0;
	}
	return 1;
}

/* LOCAL FUNCTIONS: */

/* returns WHITE if x>0, BLACK if x>0, NONE if x==0
 */
int color(int x)
{
	if(x>0) return WHITE;
	else if(x<0) return BLACK;
	else return NONE;
}

/* manage_occured() files given position into occured positions and returns DRAW if 
 * it has occured three times, else returns PROCEEDING  
 */
Game_state manage_occured(Position *position, 
                          Occured_position *occured_positions,
                          Move *last_move)
{ 
	int i;

	switch (last_move->move_type){
	case PAWN_STEP:
	case PAWN_JUMP:
	case KING_JUMP:
	case PROMOTION:
	case JUMP_PROMOTION: /* in all of this cases the position has changed unreturnably */
		realloc(occured_positions, MIN_OCCURED * sizeof(Occured_position));
		occured_positions[0].position= *position;
		occured_positions[1].frequency= -1; /* block */ 
		return PROCEEDING;
	case KING_STEP:
		for (i=0; ; i++){
			if(occured_positions[i].frequency == -1){
				if((i+1) >= sizeof(occured_positions))
					realloc(occured_positions, 2 * (i+1) * sizeof(Occured_position));
				occured_positions[i].frequency= 1;
				occured_positions[i].position= *position;
				occured_positions[i+1].frequency= -1;
				return PROCEEDING;
			}
			if(eq_positions(occured_positions[i].position, *position)){
				occured_positions[i].frequency++;
				if(occured_positions[i].frequency>=3){
					return DRAW_RP;
				}
				else {
					return PROCEEDING;
				}
			}
		} 
		assert(1==0); /* this will never happen */
		return PROCEEDING;
	default:
		assert(1==0); /* this will never happen */
		return PROCEEDING;
	} /* switch */

} /* manage_occured() */

void shift(Direction direction, int by, int *x, int *y)
{
	switch(direction){
	case NW: *x-= by;  *y+= by;  break;
	case NE: *x+= by;  *y+= by;  break;
	case SW: *x-= by;  *y-= by;  break;
	case SE: *x+= by;  *y-= by;  break;
	} 
}

/* add_moves() adds moves from `moves1' to `moves'. Only stores moves with 
 * maximal jump_length. Maintains information of allocated memory for moves and 
 * keeps block next to the moves. 
 */
void add_moves(Moves *moves, Moves moves1)
{
	int i, allocate;
	
	assert(moves!=NULL);
	if(moves->list==NULL){ 
		*moves= moves1;  
		return; 
	}
	if(moves1.list==NULL) return;
	if(moves->list[0].jump_length > moves1.list[0].jump_length){
		free(moves1.list);
		return;
	}
	if(moves->list[0].jump_length < moves1.list[0].jump_length){
		free(moves->list);
		*moves= moves1;
		return;
	}
	assert(moves->list[0].jump_length == moves1.list[0].jump_length);
	allocate= moves->allocated;
	while(allocate < moves->count + moves1.count)
		allocate*= 2;
	if(allocate > moves->allocated){
		moves->allocated= allocate;
		moves->list= realloc(moves->list, allocate * sizeof(Move));
		assert(moves->list);
	}
	for(i=0; i<moves1.count; i++)
		moves->list[moves->count+i]= moves1.list[i];
	moves->count+= moves1.count;
	free(moves1.list);
} /* end of add_moves() */

/* extend_jump() extends first part of jump by given extensions
 */
Moves extend_jump(Move first_part, Moves extensions)
{
	int i, j;
	Moves moves;

	assert(correct_coords(first_part.start.x, first_part.start.y));
	assert(correct_coords(first_part.destination.x, first_part.destination.y));

	if (extensions.count==0){
		if(extensions.allocated>0) free(extensions.list);
		first_part.jump_length= 1;
		moves.list= (Move *)malloc(8 * sizeof(Move));
		moves.list[0]= first_part;
		moves.count= 1;
		moves.allocated= 8;
		return moves;
	}
	for(i=0; i < extensions.count; i++){
		assert(correct_coords(extensions.list[i].start.x, extensions.list[i].start.y));
		assert(correct_coords(extensions.list[i].destination.x, extensions.list[i].destination.y));
		j= extensions.list[i].jump_length;
		extensions.list[i].middle[j-1]= extensions.list[i].start;
		extensions.list[i].captured_coords[j]= first_part.captured_coords[0];
		extensions.list[i].captured_pieces[j]= first_part.captured_pieces[0];
		extensions.list[i].start= first_part.start;
		extensions.list[i].jump_length++;
	}	
	return extensions;
}

/* king_jump() returns array of possible jumps of given king-piece in given direction 
 * (NULL if none found)
 */
Moves king_jump(Direction direction, const Position *position, int start_x, int start_y)
{
	Moves moves;
	Moves extensions;
	Moves moves_to_add;
	Move first_parts[8]; //maximum of first parts count is 8
	Piece captured_piece;
	Position new_pos= *position;
	int dest_x, 
		dest_y,
		captured_x= start_x,
		captured_y= start_y,
		i, j;
	moves.list= NULL;
	moves.count= 0;
	moves.allocated= 0;

	shift(direction, 1, &captured_x, &captured_y);
	if(is_out_of_board(captured_x, captured_y))
		return moves;
	while(position->placement[captured_x][captured_y] == NONE){
		shift(direction, 1, &captured_x, &captured_y);
		if(is_out_of_board(captured_x, captured_y))
			return moves;
	}
	if( (position->placement[start_x][start_y]
	     * position->placement[captured_x][captured_y]) >= 0)
		return moves;  // illegal piece to capture
	dest_x= captured_x;  dest_y= captured_y;
	shift(direction, 1, &dest_x, &dest_y);
	for( i=0 ; 
	     !is_out_of_board(dest_x, dest_y) && (position->placement[dest_x][dest_y] == NONE); 
	     i++, shift(direction, 1, &dest_x, &dest_y) 
	    ){
		assert(i<8);
		first_parts[i].start.x= start_x;
		first_parts[i].start.y= start_y;
		first_parts[i].captured_coords[0].x= captured_x;
		first_parts[i].captured_coords[0].y= captured_y;
		first_parts[i].captured_pieces[0]= 
			position->placement[captured_x][captured_y];
		first_parts[i].destination.x= dest_x;
		first_parts[i].destination.y= dest_y;
		first_parts[i].jump_length= 1;
		first_parts[i].move_type= KING_JUMP;
	}
	for(j=0; j<i; j++){
		do_move(&new_pos, first_parts[j]);
		new_pos.onmove*= -1; //keep the same player on move
		//tricky secure that king will not capture one piece twice
		captured_piece= new_pos.placement[captured_x][captured_y];
		new_pos.placement[captured_x][captured_y]=  
			new_pos.placement[start_x][start_y];
		//find jump extensions recursively
		extensions= king_jumps(&new_pos, 
		                       first_parts[j].destination.x, 
		                       first_parts[j].destination.y);
		moves_to_add= extend_jump(first_parts[j], extensions);
		add_moves(&moves, moves_to_add);
		undo_move(&new_pos, first_parts[j]);
		new_pos.onmove*= -1;
		assert(eq_positions(new_pos, *position));
	}
	return moves; 
}

/* king_step() returns array of possible steps of given king-piece in given direction
 * (NULL if none found)
 */
Moves king_step(Direction direction, const Position *position, int start_x, int start_y)
{ 
	int dest_x= start_x, 
		dest_y= start_y,
		i;
	Moves king_steps;

	king_steps.allocated= 16; // 9 would be enough
	king_steps.list= (Move *)malloc(king_steps.allocated * sizeof(Move)); 
	king_steps.count= 0;
	
	shift(direction, 1, &dest_x, &dest_y);
	for(i=0; !is_out_of_board(dest_x, dest_y) 
		      && (position->placement[dest_x][dest_y] == NONE); i++){
		king_steps.list[i].start.x= start_x;
		king_steps.list[i].start.y= start_y;
		king_steps.list[i].destination.x= dest_x;
		king_steps.list[i].destination.y= dest_y;
		king_steps.list[i].jump_length= 0;
		king_steps.list[i].move_type= KING_STEP;
		king_steps.count++;
		shift(direction, 1, &dest_x, &dest_y);
	}
	return king_steps; 
}

/* pawn_jump() returns possible jumps of given pawn-piece in given direction
 * (NULL if none found)
 */
Moves pawn_jump(Direction direction, const Position *position, int start_x, int start_y)
{
	Moves moves;
	Move first_part;
	Position new_pos= *position;
	int dest_x= start_x, 
		dest_y= start_y,
		captured_x= start_x,
		captured_y= start_y,
		i;
	moves.allocated= 0;
	moves.count= 0;
	moves.list= NULL;

	shift(direction, 1, &captured_x, &captured_y);
	shift(direction, 2, &dest_x, &dest_y);
	if(is_out_of_board(dest_x, dest_y))
		return moves;
	if( (position->placement[start_x][start_y]
	     * position->placement[captured_x][captured_y]) >= 0)
		return moves;  // illegal piece to capture
	if(position->placement[dest_x][dest_y] != NONE)
		return moves;

	assert(correct_coords(start_x, start_y));
	assert(correct_coords(captured_x, captured_y));
	assert(correct_coords(dest_x, dest_y));
	
	first_part.start.x= start_x;
	first_part.start.y= start_y;
	first_part.destination.x= dest_x;
	first_part.destination.y= dest_y;
	first_part.jump_length= 1;
	first_part.move_type= PAWN_JUMP;
	first_part.captured_coords[0].x= captured_x;
	first_part.captured_coords[0].y= captured_y;
	first_part.captured_pieces[0]= position->placement[captured_x][captured_y];
	
	// find extensions recursively:
	do_move(&new_pos, first_part); 
	new_pos.onmove= -new_pos.onmove; // keep the same player on move
	moves= extend_jump(first_part, pawn_jumps(&new_pos, dest_x, dest_y)); 

	for(i=0; i<moves.count; i++) 
		if(is_promoting(moves.list[i].destination.y, 
		   position->placement[start_x][start_y]))
			moves.list[i].move_type= JUMP_PROMOTION;
		else
			moves.list[i].move_type= PAWN_JUMP;
	return moves;
}

/* pawn_step() returns moves structure with step of given pawn-piece in given direction
 * (empty moves structure if it is illegal)
 */
Moves pawn_step(Direction direction, const Position *position, int start_x, int start_y)
{
	Moves moves;
	int dest_x= start_x, 
		dest_y= start_y;
	moves.list= NULL;
	moves.allocated= 0;
	moves.count= 0;

	if(    ((direction==NW) || (direction==NE)) 
	    && is_black_piece(position->placement[start_x][start_y]) )
		return moves;
	if(    ((direction==SW) || (direction==SE)) 
	    && is_white_piece(position->placement[start_x][start_y]) )
		return moves;
	shift(direction, 1, &dest_x, &dest_y);
	if(is_out_of_board(dest_x, dest_y))
		return moves;
	if((position->placement[dest_x][dest_y])!=NONE)
		return moves;

	moves.allocated= 8;
	moves.list= (Move *)malloc(moves.allocated * sizeof(Move)); 
	moves.list[0].start.x= start_x;
	moves.list[0].start.y= start_y;
	moves.list[0].destination.x= dest_x;
	moves.list[0].destination.y= dest_y;
	moves.list[0].jump_length= 0;
	if(is_promoting(dest_y, color(position->placement[start_x][start_y])))
		moves.list[0].move_type= PROMOTION;
	else
		moves.list[0].move_type= PAWN_STEP;
	moves.count= 1;
	return moves;
}

/* pawn_steps() returns array of all possible steps of given pawn-piece
 */
Moves pawn_steps(const Position *position, int x, int y)
{
	Moves pawn_steps;
	pawn_steps.allocated= 0;
	pawn_steps.list= NULL;
	pawn_steps.count= 0;

	add_moves(&pawn_steps, pawn_step(NW, position, x, y));
	add_moves(&pawn_steps, pawn_step(NE, position, x, y));
	add_moves(&pawn_steps, pawn_step(SW, position, x, y));
	add_moves(&pawn_steps, pawn_step(SE, position, x, y));
	return pawn_steps;
}

/* pawn_jumps() returns array of all possible jumps of given pawn-piece
 */
Moves pawn_jumps(const Position *position, int x, int y)
{
	Moves pawn_jumps;
	pawn_jumps.allocated= 0;
	pawn_jumps.list= NULL;
	pawn_jumps.count= 0;

	add_moves(&pawn_jumps, pawn_jump(NW, position, x, y));
	add_moves(&pawn_jumps, pawn_jump(NE, position, x, y));
	add_moves(&pawn_jumps, pawn_jump(SW, position, x, y));
	add_moves(&pawn_jumps, pawn_jump(SE, position, x, y));
	return pawn_jumps;
}

/* king_steps() returns array of all possible steps of given king-piece
 */
Moves king_steps(const Position *position, int x, int y)
{
	Moves king_steps;
	king_steps.allocated= 0;
	king_steps.list= NULL;
	king_steps.count= 0;

	add_moves(&king_steps, king_step(NW, position, x, y));
	add_moves(&king_steps, king_step(NE, position, x, y));
	add_moves(&king_steps, king_step(SW, position, x, y));
	add_moves(&king_steps, king_step(SE, position, x, y));
	return king_steps;
}

/* king_jumps() returns array of all possible jumps of given king-piece
 */
Moves king_jumps(const Position *position, int x, int y)
{
	Moves king_jumps;
	king_jumps.allocated= 0;
	king_jumps.list= NULL;
	king_jumps.count= 0;

	add_moves(&king_jumps, king_jump(NW, position, x, y));
	add_moves(&king_jumps, king_jump(NE, position, x, y));
	add_moves(&king_jumps, king_jump(SW, position, x, y));
	add_moves(&king_jumps, king_jump(SE, position, x, y));
	return king_jumps;
}

Moves get_piece_moves(const Position *position, int x, int y)
{
	Moves piece_moves;
	piece_moves.list= NULL;
	piece_moves.allocated= 0;
	piece_moves.count= 0;

	switch(position->placement[x][y]){
	case NONE:
		break;
	case BLACK_PAWN:
	case WHITE_PAWN:
		if(position->onmove * position->placement[x][y] < 0) break; /* other player on move */
		add_moves(&piece_moves, pawn_jumps(position, x, y));
		if( piece_moves.count==0 )
			add_moves(&piece_moves, pawn_steps(position, x, y));
		break;
	case BLACK_KING:
	case WHITE_KING:
		if(position->onmove * position->placement[x][y] < 0) break; /* other player on move */
		add_moves(&piece_moves, king_jumps(position, x, y));
		if( piece_moves.count==0 )
			add_moves(&piece_moves, king_steps(position, x, y));
		break;
	}
	return piece_moves;
}
